import sys
import random
import datetime
import time
from dateutil import parser
from algo_trader.clients import BitmexOrder
import requests
import os
import signal

from socket import gaierror


class Lotus:
    def __init__(self, client, symbols, settings, setting_file_name):
        self.client = client
        self.symbols = symbols
        self.settings = settings
        if self.client.testnet:
            from algo_trader.settings import TESTNET_LOTUS_LINK
            self._api_endpoint = TESTNET_LOTUS_LINK
        else:
            from algo_trader.settings import LIVE_LOTUS_LINK
            self._api_endpoint = LIVE_LOTUS_LINK
        self._settings_path = os.path.abspath(os.path.join(os.path.dirname(
            os.path.abspath(__file__)), os.pardir, 'settings', setting_file_name))

        self._last_signal_timestamp = self.get_api_signal_retry()['timestamp']

        self._last_timestamp = {}
        self._last_entry = {}
        self._last_sl = {}

        self.me_count = 0  # initialize manage entries count for candle check every 3rd time

        signal.signal(signal.SIGINT, self.signal_handler)

        for symbol in symbols:
            self._last_timestamp[symbol] = datetime.datetime.utcnow()
            self._last_entry[symbol] = ''
            self._last_sl[symbol] = ''

        if not self.client.is_connected:
            print(
                "Invalid API ID or API Secret, please restart and provide the right keys", flush=True)
            sys.exit()

    # helper function
    def ceil_dt(self, dt, delta):
        return dt + (datetime.datetime.min - dt) % delta

    def signal_handler(self, signal, frame):
        print('Doing things before exit...')
        for symbol in self.symbols:
            if self.order.props[symbol]['wait_stop']:
                self.order.close(symbol)  # cancel old order first
        sys.exit(0)

    def is_new_signal_symbol(self, signal, symbol):
        timestamp = parser.parse(signal[symbol]['timestamp'])
        if self._last_timestamp[symbol] != timestamp:
            self._last_timestamp[symbol] = timestamp
            if self._last_entry[symbol] != signal[symbol]['entry'] or self._last_sl[symbol] != signal[symbol]['stoploss']:
                self._last_entry[symbol] = signal[symbol]['entry']
                self._last_sl[symbol] = signal[symbol]['stoploss']
                return True

    def get_api_signal(self):
        auth = {'token': self.settings.token}
        try:
            return requests.get(self._api_endpoint, params=auth).json()
        except ValueError:
            print("Can't get signal on endpoint {} with token {}."
                  "It could be that you are timed out because you try to access with multiple IP's.\nIn this case wait 40 minutes to be able to access again.\(Don't need to restart)".format(
                      self._api_endpoint, self.settings.token))
            return False
        except (requests.exceptions.Timeout, requests.exceptions.ConnectionError, gaierror):
            time.sleep(30)
            print('Retrying connection to endpoint')
            return self.get_api_signal()
        except requests.exceptions.TooManyRedirects:
            print('Bad URL. Try to setup the correct endpoint. Current endpoint: {}'.format(
                self._api_endpoint))
            return False
        except requests.exceptions.RequestException as e:
            print('Another exception occured {}'.format(e))
            return False

    def get_api_signal_retry(self):
        while True:
            signal = self.get_api_signal()
            if signal:
                return signal
            else:
                time.sleep(10)

    def is_new_signal(self, signal):
        # tests if signal has new timestamp
        if signal['timestamp'] != self._last_signal_timestamp:
            self._last_signal_timestamp = signal['timestamp']
            return True
        else:
            return False

    def rand_sec(self, start=2, end=6):
        return random.randrange(start, end)

    def check_modify(self, symbol, sl):
        res = self.order.modifiy_stop(symbol, sl)
        if type(res) is str:
            if 'noValChanged' not in res:
                print('Modify order Error {}:'.format(symbol), res, flush=True)
        else:
            print('Modifying order in {} to new stoploss: {}'.format(
                symbol, sl), flush=True)
            self.order.SL = sl

    def manage_open_stop_orders(self, symbol):
        if self.order.props[symbol]['wait_stop']:
            print("Cancel old order {} first.".format(symbol))
            self.order.stoporder_cancel(symbol)
            time.sleep(3)
        elif self.order.props[symbol]['open']:
            print("Manage open stop orders, order open. Trying to close.", flush=True)
            self.order.close(symbol, -self.client.open_contracts(symbol))
            time.sleep(3)

    def process_signals(self, ns, start=False):
        for symbol in self.symbols:
            if start:
                if ns[symbol]['signal']:
                    self.order.check_open_bot_order(
                        symbol, ns[symbol]['entry'], ns[symbol]['stoploss'])
                    time.sleep(3)
            signal = ns[symbol]['signal']
            if signal:
                if self.is_new_signal_symbol(ns, symbol):
                    entry = ns[symbol]['entry']
                    stoploss = ns[symbol]['stoploss']
                    if signal in [-1, 1]:  # signal has open position
                        if signal == 1:
                            if self.order.props[symbol]['qty'] > 0 or self.client.open_contracts(symbol) > 0:
                                self.check_modify(symbol, stoploss)
                            elif self.client.open_contracts(symbol) < 0:
                                print('Closing {} order.'.format(
                                    symbol), flush=True)
                                close = self.order.close(
                                    symbol, qty=-self.order.props[symbol]['qty'])
                                if close:
                                    print('Closed open Bot old short position in {}, because now the Bot\'s Position is long.'.format(
                                        symbol), flush=True)
                                else:
                                    print('Error closing {} order.'.format(
                                        symbol), flush=True)
                            else:
                                cp = self.client.last_current_price(symbol)
                                if cp < entry:
                                    sl_dist = abs(stoploss-entry)
                                    sl_new_dist = abs(cp - stoploss)
                                    if sl_new_dist > 0.15 * sl_dist:
                                        time.sleep(2)
                                        contracts = self.order.calc_pos_size(
                                            symbol, sl_new_dist)
                                        print("Open market order long into open Signal. {} {} Contracts, Stoploss: {}".format(
                                            symbol, contracts, stoploss), flush=True)
                                        self.order.bracket_market_order(
                                            symbol, contracts, stoploss, cp)
                        else:
                            if self.order.props[symbol]['qty'] < 0 or self.client.open_contracts(symbol) < 0:
                                self.check_modify(symbol, stoploss)
                            elif self.client.open_contracts(symbol) > 0:
                                print('Closing order on {}.'.format(
                                    symbol), flush=True)
                                close = self.order.close(
                                    symbol, qty=-self.order.props[symbol]['qty'])
                                if close:
                                    print('Closed open Bot old long position in {}, because now the Bot\'s Position is short.'.format(
                                        symbol), flush=True)
                                else:
                                    print('Error closing {} order.'.format(
                                        symbol), flush=True)
                            else:
                                cp = self.client.last_current_price(symbol)
                                if cp > entry:
                                    sl_dist = abs(stoploss-entry)
                                    sl_new_dist = abs(cp - stoploss)
                                    if sl_new_dist > 0.15 * sl_dist:
                                        time.sleep(2)
                                        contracts = self.order.calc_pos_size(
                                            symbol, sl_new_dist)
                                        print("Open market order short into open Signal. {} {} Contracts, Stoploss: {}".format(
                                            symbol, contracts, stoploss), flush=True)
                                        self.order.bracket_market_order(
                                            symbol, -contracts, stoploss, cp)
                    elif signal in [-2, 2]:  # signal has pending order
                        self.manage_open_stop_orders(symbol)
                        contracts = self.order.calc_pos_size(
                            symbol, abs(entry - stoploss))
                        if signal == 2:
                            print("Open bracket stop order long. {} {} Contracts, Entry: {}, Stoploss: {}".format(
                                symbol, contracts, entry, stoploss), flush=True)
                            self.order.bracket_stop_order(
                                symbol, contracts, entry, stoploss)
                        else:
                            print("Open bracket stop order short. {} {} Contracts, Entry: {}, Stoploss: {}".format(
                                symbol, contracts, entry, stoploss), flush=True)
                            self.order.bracket_stop_order(
                                symbol, -contracts, entry, stoploss)
                    else:
                        if self.order.props[symbol]['open'] or self.client.open_contracts(symbol) > 0:
                            self.order.close(symbol)
            elif signal != 0:
                self.manage_open_stop_orders(symbol)
                print("This position update in {} is only for premium members. Next update: {}".format(symbol,
                                                                                                       self.ceil_dt(datetime.datetime.now(), datetime.timedelta(minutes=30))), flush=True)

    def run(self):
        # Main loop
        self.order = BitmexOrder(self.symbols,
                                 self.client, self.settings, self._settings_path)
        while True:
            signal = self.get_api_signal()
            if not signal:
                print("Trying again in 1 Minute.", flush=True)
                time.sleep(60)
            else:
                self.process_signals(signal, start=True)
                print(
                    "BOT IS RUNNING... **(ignore potential bravado.core warning)**", flush=True)
                break
        while True:
            time.sleep(10)
            self.me_count += 1
            self.order.manage_entries(self.symbols, me_count=self.me_count)
            now = datetime.datetime.utcnow()
            if now.minute % 60 == 0:
                exec_minute = now.minute
                time.sleep(19 + self.rand_sec(start=1, end=3))
                count_s_tries = 0
                while True:
                    signal = self.get_api_signal()
                    if not signal:
                        # in case the receiving of signal is not authorized
                        break
                    elif self.is_new_signal(signal):
                        # make sure to get the signal
                        self.process_signals(signal)
                        break
                    count_s_tries += 1
                    if count_s_tries > 22:
                        print('No new signal.', flush=True)
                        break
                    time.sleep(self.rand_sec())
                while True:
                    time.sleep(10)
                    self.order.manage_entries(
                        self.symbols, me_count=self.me_count)
                    if datetime.datetime.utcnow().minute != exec_minute:
                        break



import json
import math
import bitmex
import time
import uuid
import pickle
import sys

from bravado.exception import HTTPBadGateway, HTTPUnauthorized, HTTPBadRequest, HTTPNotFound, HTTPGatewayTimeout, HTTPServiceUnavailable, HTTPTooManyRequests, HTTPServerError


class BitmexClient:
    def __init__(self, api_key=None, api_secret=None, testnet=True):
        self.client = bitmex.bitmex(
            api_key=api_key, api_secret=api_secret, test=testnet)
        self._last_wallet_balance = ''
        self._last_currentprice = {}
        self.testnet = testnet

    def __call__(self):
        return self.client

    def get_margin(self, prop):
        try:
            wallet_balance = self.client.User.User_getMargin().result()[
                0][str(prop)]
            self._last_wallet_balance = wallet_balance
            return wallet_balance
        except (HTTPBadRequest, HTTPGatewayTimeout):
            return self._last_wallet_balance

    def get_position(self, symbol, prop):
        position = 0
        i = 0
        while True:
            try:
                position = self.client.Position.Position_get(
                    filter=json.dumps({'symbol': symbol})).result()[0][0][str(prop)]
                break
            except (IndexError, HTTPBadRequest, HTTPBadGateway) as e:
                if 'expired' in str(e):
                    print('Getting open contracts in {} failed. Expired error.'.format(
                        symbol), flush=True)
                else:
                    time.sleep(2)  # try again after 2 seconds
                    print("Can't get position in {}. Trying again...".format(symbol), e)
                    i += 1
                    if i > 1:
                        print("Can't get position in {} for 2nd time.".format(symbol))
                        return False
        return position

    def get_histories(self, symbols=['XBTUSD'], binSize='5m', count=15):
        histories = {}
        for symbol in symbols:
            i = 0
            while True:
                try:
                    histories[symbol] = self.client.Trade.Trade_getBucketed(symbol=symbol,
                                                                            binSize=binSize,
                                                                            count=count,
                                                                            reverse=True,
                                                                            ).result()
                    break
                except HTTPBadRequest as e:
                    if 'expired' in str(e):
                        print('Getting {} history failed. Expired error.'.format(
                            symbol), flush=True)
                    else:
                        time.sleep(2)  # try again after 2 seconds
                        print("Can't get history. Trying again...", e)
                        i += 1
                        if i > 1:
                            print("Can't get history for 2nd time.")
                            break
        return histories

    def get_current_price(self, symbol):
        try:
            price = self.client.Trade.Trade_get(symbol=symbol,
                                                count=2,
                                                reverse=True,
                                                ).result()[0][0]['price']
            self._last_currentprice[symbol] = price
            return price
        except (IndexError, HTTPBadRequest, HTTPGatewayTimeout, HTTPBadGateway, HTTPServiceUnavailable, HTTPServerError) as e:
            if 'expired' in str(e):
                print('Current price expired. Returning last price.', flush=True)
            elif 'Service Unavailable' in str(e):
                print('Server maintenance. Trying to reconnect in 5 Minutes...')
                while True:
                    time.sleep(360)
                    current_price = self.get_current_price(symbol)
                    if current_price:
                        break
            # else:
            #     # print('API error getting current price', e, flush=True)
            #     pass
            return self.last_current_price(symbol)

    def get_current_price_candle(self, symbol):
        try:
            price = self.client.Trade.Trade_getBucketed(symbol=symbol,
                                                        binSize='1m',
                                                        count=1,
                                                        reverse=True,
                                                        ).result()[0][0]
            return price
        except (IndexError, HTTPBadRequest, HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable, HTTPTooManyRequests) as e:
            if 'Service Unavailable' in str(e):
                print(
                    'Server maintenance. Trying to reconnect in 5 Minutes...', flush=True)
                while True:
                    time.sleep(360)
                    current_price = self.get_current_price(symbol)
                    if current_price:
                        break
            return self._last_currentprice[symbol]

    def last_current_price(self, symbol):
        if symbol in self._last_currentprice.keys():
            return self._last_currentprice[symbol]
        else:
            time.sleep(2)
            return self.get_current_price(symbol)

    def unrealised_pnl(self, symbol):
        return self.get_position(symbol, 'unrealisedPnl')

    def open_contracts(self, symbol):
        res = self.get_position(symbol, 'currentQty')
        if res is not False:
            return res
        else:
            return 0

    def current_price_position(self, symbol):
        return self.get_position(symbol, 'lastPrice')

    @property
    def is_connected(self):
        if not self.acc_balance:
            return False
        else:
            return True

    @property
    def acc_balance(self):
        try:
            return self.get_margin('walletBalance')
        except HTTPUnauthorized:
            return False


class BitmexOrder:
    def __init__(self, symbols, client, settings, settings_path, magic='algo-trader'):
        self.client = client
        self.settings = settings
        self._settings_path = settings_path
        self.magic = magic
        self._symbols = symbols

        self.props = {}
        for symbol in symbols:
            self.props[symbol] = {
                'stoploss_id': '', 'stop_id': '', 'entry': .0, 'SL': .0, 'TP': .0, 'open': False, 'wait_stop': False, 'qty': 0, 'initialSL': .0}

        self.show_slippage = False  # only for testing
        if self.show_slippage:
            self.entry_acc_balance = {}
            self.between_profits = {}
            for symbol in symbols:
                self.between_profits[symbol] = .0
                self.entry_acc_balance[symbol] = .0

    def check_open_bot_order(self, symbol, entry, sl):
        settings = pickle.load(open(self._settings_path, 'rb'))
        self.props[symbol]['stoploss_id'] = settings.symbols[symbol]['last_order_stoploss_id']
        self.props[symbol]['stop_id'] = settings.symbols[symbol]['last_order_stop_id']
        self.props[symbol]['SL'] = settings.symbols[symbol]['last_order_SL']
        self.props[symbol]['qty'] = settings.symbols[symbol]['last_order_qty']
        if self.props[symbol]['stoploss_id']:
            try:
                res = self.client.client.Order.Order_getOrders(
                    filter=json.dumps({"symbol": symbol, "open": True})).result()
                if res[0] and self.props[symbol]['stoploss_id'] in [order['clOrdID'] for order in res[0]]:
                    print("Bot order in {} open, stoploss might be modified".format(
                        symbol), flush=True)
                    self.modifiy_stop(symbol, sl)
                    self.props[symbol]['entry'] = entry
                    self.props[symbol]['open'] = True
                    time.sleep(2)
            except HTTPBadRequest as e:
                print("Error getting orders.", e)
        if self.props[symbol]['stop_id']:
            try:
                res = self.client.client.Order.Order_getOrders(
                    filter=json.dumps({"symbol": symbol, "open": True})).result()
                if res[0] and self.props[symbol]['stop_id'] in [order['clOrdID'] for order in res[0]]:
                    print("Bot order in {} open, stop will be deleted.".format(
                        symbol), flush=True)
                    self.stoporder_cancel(symbol)
                    time.sleep(2)
            except HTTPBadRequest as e:
                print("Error getting orders.", e)

    def save_open_id(self, symbol, id, stop_type='normal'):
        settings = pickle.load(open(self._settings_path, 'rb'))
        if stop_type == 'normal':
            self.props[symbol]['stop_id'] = id
            settings.symbols[symbol]['last_order_stop_id'] = id
            settings.symbols[symbol]['last_order_qty'] = self.props[symbol]['qty']
            pickle.dump(settings, file=open(self._settings_path, 'wb'))
        elif stop_type == 'stoploss':  # also save the stoploss
            self.props[symbol]['stoploss_id'] = id
            settings.symbols[symbol]['last_order_stoploss_id'] = id
            settings.symbols[symbol]['last_order_SL'] = self.props[symbol]['SL']
            pickle.dump(settings, file=open(self._settings_path, 'wb'))
        else:
            print("No Stop Type given.")

    def set_position_open(self, symbol, dir):
        self.props[symbol]['wait_stop'] = False
        self.props[symbol]['open'] = True
        print("{} Position has been opened in {}, amending Stoploss to".format(dir, symbol),
              self.props[symbol]['SL'], flush=True)
        self.stoploss_order(symbol)
        self.calculate_between_profits(symbol)

    def calculate_between_profits(self, symbol):
        between_profs = .0
        for sym in self._symbols:
            if symbol != sym:
                between_profs += self.between_profits[sym]
                self.between_profits[sym] = .0
        return between_profs

    def set_position_closed(self, symbol, dir, currentPrice=False):
        self.props[symbol]['open'] = False
        pnl_points = self.price_decimals(
            symbol, self.props[symbol]['SL']-self.props[symbol]['entry'])
        qty = self.props[symbol]['qty']
        if dir == "Short":
            pnl_points = -pnl_points
            qty = -qty
        if currentPrice:
            exitPrice = self.client.get_current_price(symbol)
        else:
            exitPrice = self.props[symbol]['SL']
        print("{} Position {} closed. {} Contracts from {} to {}. PnL in Points: {}".format(symbol,
                                                                                            dir, qty,
                                                                                            self.props[symbol]['entry'],
                                                                                            exitPrice,
                                                                                            pnl_points), flush=True)
        if self.show_slippage and self.props[symbol]['initialSL']:
            acc_pos_size = float(
                self.settings.symbols[symbol]['position_size'])/100

            current_acc_balance = float(self.client.acc_balance)

            between_profs = self.calculate_between_profits(symbol)

            # exit account balance - current account balance
            acc_balance_diff = current_acc_balance - \
                between_profs - self.entry_acc_balance[symbol]
            print("current_acc_balance", current_acc_balance, "between_profs", between_profs,
                  "self.entry_acc_balance[symbol]", self.entry_acc_balance[symbol])

            # in acc balance difference
            self.between_profits[symbol] += acc_balance_diff

            initial_risk_points = abs(
                self.props[symbol]['initialSL'] - self.props[symbol]['entry'])
            pnl_rr = pnl_points / initial_risk_points

            real_gain = (acc_balance_diff / self.entry_acc_balance[symbol])
            expected_gain = pnl_rr * acc_pos_size

            # slippage is real acc balance change relative - executed pnl percent
            print("expected_gain", expected_gain,
                  "real_gain", real_gain)
            print("Slippage: {}%".format(
                round((real_gain - expected_gain) / abs(expected_gain) * 100, 2)), flush=True)
        self.props[symbol]['qty'] = 0

    def has_position_opened(self, symbol, cp, candlePrice):
        if self.client.open_contracts(symbol) != 0:
            if self.props[symbol]['qty'] > 0:
                _cp = candlePrice['high'] if candlePrice else cp
                if _cp >= self.props[symbol]['entry']:
                    self.set_position_open(symbol, 'Long')
            elif self.props[symbol]['qty'] < 0:
                _cp = candlePrice['low'] if candlePrice else cp
                if _cp <= self.props[symbol]['entry']:
                    self.set_position_open(symbol, 'Short')

    def has_position_closed(self, symbol, cp, candlePrice):
        # wait for if position really closed and prevent bad requests
        time.sleep(1)
        if self.client.open_contracts(symbol) == 0:
            if self.props[symbol]['qty'] > 0:
                _cp = candlePrice['low'] if candlePrice else cp
                if _cp <= self.props[symbol]['SL']:
                    self.set_position_closed(symbol, 'Long')
            elif self.props[symbol]['qty'] < 0:
                _cp = candlePrice['low'] if candlePrice else cp
                if _cp >= self.props[symbol]['SL']:
                    self.set_position_closed(symbol, 'Short')

    def manage_entries(self, symbols, me_count=0, candlePrice=''):
        for symbol in symbols:
            if not candlePrice:
                cp = self.client.get_current_price(symbol)
            else:
                cp = ''
            time.sleep(0.5)
            if self.props[symbol]['wait_stop']:
                self.has_position_opened(symbol, cp, candlePrice)
            elif self.props[symbol]['open']:
                self.has_position_closed(symbol, cp, candlePrice)

        # every 30 seconds additional candle check to not miss entry/exit
        if me_count and me_count % 3 == 0:
            if self.props[symbol]['qty'] > 0:
                candlePrice = self.client.get_current_price_candle(symbol)
                # + 1 to prevent endless loop
                self.manage_entries(self._symbols, candlePrice=candlePrice)
            elif self.props[symbol]['qty'] < 0:
                candlePrice = self.client.get_current_price_candle(symbol)
                self.manage_entries(self._symbols, candlePrice=candlePrice)

    def generate_id(self, symbol, _type='stop'):
        return "{}_{}_{}_{}".format(self.magic, symbol, str(uuid.uuid4().fields[-1])[:5], _type)

    def price_decimals(self, symbol, price):
        if symbol == 'XBTUSD':
            return round(price * 2) / 2
        elif symbol == 'ETHUSD':
            return round(price * 20) / 20
        else:
            print('{} is not a valid symbol for price decimal rounding.'.format(symbol))
            sys.exit()

    def calc_pos_size(self, symbol, sl_distance):
        acc_balance = float(self.client.acc_balance)
        if self.show_slippage:
            self.entry_acc_balance[symbol] = acc_balance
        ps = float(self.settings.symbols[symbol]['position_size'])
        if symbol == 'XBTUSD':
            btc_usd = self.client.last_current_price(symbol)
            lot_step = 100
            riskvalue = (acc_balance / 100000000) * \
                float(btc_usd) * (ps/100)  # riskvalue in usd
            pos_size = int(
                round((math.floor(riskvalue / (sl_distance / btc_usd)) / lot_step)) * lot_step)
        elif symbol == 'ETHUSD':
            tickvalue = 100
            lot_step = 1
            riskvalue = acc_balance * (ps/100)
            pos_size = int(
                round((math.floor(riskvalue / (sl_distance * tickvalue)) / lot_step)) * lot_step)
        else:
            print('Symbol not supported')
            sys.exit()
        if pos_size < lot_step:
            return lot_step
        else:
            return pos_size

    def stoploss_order(self, symbol, execPrice='LastPrice'):
        try:
            self.props[symbol]['stoploss_id'] = self.generate_id(
                symbol, _type='SL')
            self.save_open_id(
                symbol, self.props[symbol]['stoploss_id'], stop_type='stoploss')
            return self.client.client.Order.Order_new(
                symbol=symbol, ordType='Stop', clOrdID=self.props[symbol]['stoploss_id'], orderQty=-self.props[symbol]['qty'],
                stopPx=self.price_decimals(symbol, self.props[symbol]['SL']), execInst=execPrice
            ).result()
        except HTTPBadRequest as e:
            print("Error placing stoploss order.", e, flush=True)

    def order_cancel(self, symbol, orderId):
        try:
            order_res = self.client.client.Order.Order_cancel(
                clOrdID=orderId
            ).result()
            self.props[symbol]['wait_stop'] = False
            self.props[symbol]['qty'] = 0
            return order_res
        except (HTTPBadRequest, HTTPNotFound) as e:
            print("Error canceling order.", e, flush=True)

    def stoploss_cancel(self, symbol):
        self.order_cancel(symbol, self.props[symbol]['stoploss_id'])

    def stoporder_cancel(self, symbol):
        self.order_cancel(symbol, self.props[symbol]['stop_id'])

    def bracket_stop_order(self, symbol, orderQty, entry, sl, tp=0):
        try:
            if self.props[symbol]['qty'] != 0:
                self.close(symbol)
                time.sleep(1)
            self.props[symbol]['stop_id'] = self.generate_id(symbol)
            order_res = self.client.client.Order.Order_new(
                symbol=symbol, ordType='Stop', clOrdID=self.props[symbol]['stop_id'], orderQty=orderQty,
                stopPx=self.price_decimals(symbol, entry), execInst='LastPrice'
            ).result()
            self.props[symbol]['qty'] = orderQty
            self.save_open_id(symbol, self.props[symbol]['stop_id'])
            self.props[symbol]['wait_stop'] = True
            self.props[symbol]['SL'] = sl
            self.props[symbol]['initialSL'] = sl
            self.props[symbol]['TP'] = tp
            self.props[symbol]['entry'] = entry
            return order_res
        except HTTPBadRequest as e:
            print("Error placing bracket stop order in {}.".format(
                symbol), e, flush=True)

    def bracket_market_order(self, symbol, orderQty, sl, cp, tp=0):
        try:
            entry = self.market_order(symbol, orderQty)
            self.props[symbol]['SL'] = sl
            self.props[symbol]['initialSL'] = sl
            self.props[symbol]['TP'] = tp
            time.sleep(5)
            stoploss = self.stoploss_order(symbol)
            self.props[symbol]['qty'] = orderQty
            self.props[symbol]['entry'] = cp
            return entry, stoploss
        except HTTPBadRequest as e:
            print("Error placing bracket market order in {}.".format(
                symbol), e, flush=True)

    def market_order(self, symbol, orderQty):
        try:
            order_res = self.client.client.Order.Order_new(
                symbol=symbol, ordType='Market', orderQty=orderQty
            ).result()
            self.props[symbol]['qty'] = orderQty
            self.props[symbol]['open'] = True
            self.props[symbol]['wait_stop'] = False
            return order_res
        except HTTPBadRequest as e:
            print("Error placing market order.", e, flush=True)

    def modifiy_stop(self, symbol, newStopLoss):
        try:
            # could be clOrdID, not sure difference
            res = self.client.client.Order.Order_amend(
                origClOrdID=self.props[symbol]['stoploss_id'], stopPx=self.price_decimals(symbol,
                                                                                          newStopLoss)
            ).result()
            self.props[symbol]['SL'] = newStopLoss
            return res
        except HTTPBadRequest as e:
            if 'Invalid amend' in str(e):
                return 'noValChanged'
            elif 'Invalid origClOrdID' in str(e):
                return 'invalidID'
            else:
                return str(e)

    def close(self, symbol, qty=0):
        if not qty:
            qty = -self.props[symbol]['qty']
        if self.props[symbol]['open']:
            # note: have to delete stop too
            try:
                close = self.client.client.Order.Order_new(
                    symbol=symbol, ordType='Market', execInst='Close', orderQty=qty
                ).result()
                time.sleep(1)  # wait before making new API call
                if self.props[symbol]['qty'] > 0:
                    self.set_position_closed(
                        symbol, 'Long', currentPrice=True)
                elif self.props[symbol]['qty'] < 0:
                    self.set_position_closed(
                        symbol, 'Short', currentPrice=True)
                self.stoploss_cancel(symbol)
                return close
            except HTTPBadRequest as e:
                print("Error closing order in {}.".format(symbol), e, flush=True)
        else:
            return False



import json
import math
import time
import uuid
import pickle
import sys
from binance.client import Client as binanceClient, BinanceAPIException, BinanceRequestException, NotImplementedException

import datetime

from bravado.exception import HTTPBadGateway, HTTPUnauthorized, HTTPBadRequest, HTTPNotFound, HTTPGatewayTimeout, HTTPServiceUnavailable


class BinanceClient:
    def __init__(self, api_key=None, api_secret=None, testnet=True):
        self.client = binanceClient(
            api_key, api_secret, {"timeout": 20}, testnet=testnet)
        self._last_wallet_balance = ''
        self._last_currentprice = {}
        self.testnet = testnet

    def __call__(self):
        return self.client

    def acc_balance(self, symbol):
        if 'USDT' in symbol:
            try:
                return self.client.futures_account_balance()[1]['balance']
            except (BinanceAPIException, BinanceRequestException, NotImplementedException) as e:
                print(e.status_code)
                print(e.message)
        else:
            try:
                s_bal = symbol.replace('USD', '')  # balance to search for
                balances = self.client.futures_coin_account_balance()
                return next(sym for sym in balances if sym["asset"] == s_bal)['balance']
            except (BinanceAPIException, BinanceRequestException, NotImplementedException) as e:
                print(e.status_code)
                print(e.message)

    def get_position(self, symbol, prop):
        '''

        example return of `s_info`:
            {'symbol': 'BTCUSDT', 'positionAmt': '0.001',
            'entryPrice': '41694.23', 'markPrice': '41822.27000000', 'unRealizedProfit': '0.12804000',
            'liquidationPrice': '0', 'leverage': '20', 'maxNotionalValue': '10000000', 'marginType': 'cross',
            'isolatedMargin': '0.00000000', 'isAutoAddMargin': 'false', 'positionSide': 'BOTH',
            'notional': '41.82227000', 'isolatedWallet': '0', 'updateTime': 1632491393073}   

        '''
        try:
            s_info = self.client.futures_position_information(symbol=symbol)
            return s_info[prop]
        except (BinanceAPIException, BinanceRequestException, NotImplementedException) as e:
            print(e.status_code)
            print(e.message)

    def get_histories(self, symbols=['BTCUSDT'], count=15, interval=30):
        histories = {}
        for symbol in symbols:
            i = 0
            while True:
                try:
                    histories[symbol] = self.client.get_klines(symbol=symbol, interval=str(interval)+'m', limit=count)
                    break
                except (BinanceAPIException, BinanceRequestException, NotImplementedException) as e:
                    if 'expired' in str(e):
                        print('Getting {} history failed. Expired error.'.format(
                            symbol), flush=True)
                    else:
                        time.sleep(2)  # try again after 2 seconds
                        print("Can't get history. Trying again...", e)
                        i += 1
                        if i > 1:
                            print("Can't get history for 2nd time.")
                            break
        return histories

    def get_current_price(self, symbol):
        try:
            price = self.client.Trade.Trade_get(symbol=symbol,
                                                count=2,
                                                reverse=True,
                                                ).result()[0][0]['price']
            self._last_currentprice[symbol] = price
            return price
        except (IndexError, HTTPBadRequest, HTTPGatewayTimeout, HTTPBadGateway, HTTPServiceUnavailable) as e:
            if 'expired' in str(e):
                print('Current price expired. Returning last price.', flush=True)
            elif 'Service Unavailable' in str(e):
                print('Server maintenance. Trying to reconnect in 5 Minutes...')
                while True:
                    time.sleep(360)
                    current_price = self.get_current_price(symbol)
                    if current_price:
                        break
            # else:
            #     # print('API error getting current price', e, flush=True)
            #     pass
            return self.last_current_price(symbol)

    def last_current_price(self, symbol):
        if symbol in self._last_currentprice.keys():
            return self._last_currentprice[symbol]
        else:
            time.sleep(2)
            return self.get_current_price(symbol)

    def unrealised_pnl(self, symbol):
        return self.get_position(symbol, 'unRealizedProfit')

    def open_contracts(self, symbol):
        res = self.get_position(symbol, 'positionAmt')
        if res != '0.000':
            return res
        else:
            return 0

    def current_price_position(self, symbol):
        return self.get_position(symbol, 'markPrice')

    @property
    def is_connected(self):
        print(self.acc_balance('BTCUSDT'))
        if isinstance(self.acc_balance('BTCUSDT'), str):
            return True
        else:
            return False


class BinanceOrder(BinanceClient):
    def __init__(self, symbols, settings, settings_path, magic='algo-trader', api_key=None, api_secret=None, testnet=True):
        self.settings = settings
        self.magic = magic
        self._settings_path = settings_path

        self.props = {}
        for symbol in symbols:
            self.props[symbol] = {
                'stoploss_id': '', 'stop_id': '', 'entry': .0, 'SL': .0, 'TP': .0, 'open': False, 'wait_stop': False, 'qty': 0}

        super().__init__(api_key=api_key, api_secret=api_secret, testnet=testnet)

    def check_open_bot_order(self, symbol, entry, sl):
        settings = pickle.load(open(self._settings_path, 'rb'))
        self.props[symbol]['stoploss_id'] = settings.symbols[symbol]['last_order_stoploss_id']
        self.props[symbol]['stop_id'] = settings.symbols[symbol]['last_order_stop_id']
        self.props[symbol]['SL'] = settings.symbols[symbol]['last_order_SL']
        if self.props[symbol]['stoploss_id']:
            try:
                res = self.client.Order.Order_getOrders(
                    filter=json.dumps({"symbol": symbol, "open": True})).result()
                if res[0] and self.props[symbol]['stoploss_id'] in [order['clOrdID'] for order in res[0]]:
                    print("Bot order in {} open, stoploss might be modified".format(
                        symbol), flush=True)
                    self.modifiy_stop(symbol, sl)
                    self.props[symbol]['entry'] = entry
                    self.props[symbol]['open'] = True
                    time.sleep(2)
            except HTTPBadRequest as e:
                print("Error getting orders.", e)
        if self.props[symbol]['stop_id']:
            try:
                res = self.client.client.Order.Order_getOrders(
                    filter=json.dumps({"symbol": symbol, "open": True})).result()
                if res[0] and self.props[symbol]['stop_id'] in [order['clOrdID'] for order in res[0]]:
                    print("Bot order in {} open, stop will be deleted.".format(
                        symbol), flush=True)
                    self.stoporder_cancel(symbol)
                    time.sleep(2)
            except HTTPBadRequest as e:
                print("Error getting orders.", e)

    def save_open_id(self, symbol, id, stop_type='normal'):
        settings = pickle.load(open(self._settings_path, 'rb'))
        if stop_type == 'normal':
            self.props[symbol]['stop_id'] = id
            settings.symbols[symbol]['last_order_stop_id'] = id
            pickle.dump(settings, file=open(self._settings_path, 'wb'))
        elif stop_type == 'stoploss':
            self.props[symbol]['stoploss_id'] = id
            settings.symbols[symbol]['last_order_stoploss_id'] = id
            settings.symbols[symbol]['last_order_SL'] = self.props[symbol]['SL']
            pickle.dump(settings, file=open(self._settings_path, 'wb'))
        else:
            print("No Stop Type given.")

    def manage_entries(self, symbols):
        for symbol in symbols:
            cp = self.client.get_current_price(symbol)
            if self.props[symbol]['wait_stop']:
                time.sleep(0.5)
                if self.props[symbol]['qty'] > 0 and cp >= self.props[symbol]['entry'] and self.client.open_contracts(symbol) != 0:
                    self.props[symbol]['wait_stop'] = False
                    self.props[symbol]['open'] = True
                    print("Long Position has been opened in {}, amending Stoploss to".format(symbol),
                          self.props[symbol]['SL'], flush=True)
                    self.stoploss_order(symbol)
                elif self.props[symbol]['qty'] < 0 and cp <= self.props[symbol]['entry'] and self.client.open_contracts(symbol) != 0:
                    self.props[symbol]['wait_stop'] = False
                    self.props[symbol]['open'] = True
                    print("Short Position has been opened in {}, amending Stoploss to".format(symbol),
                          self.props[symbol]['SL'], flush=True)
                    self.stoploss_order(symbol)
            elif self.props[symbol]['open']:
                if self.props[symbol]['qty'] > 0 and cp <= self.props[symbol]['SL'] and self.client.open_contracts(symbol) == 0:
                    self.props[symbol]['open'] = False
                    print("Long Position {} closed. {} Contracts from {} to {}. PnL in Points: {}".format(symbol,
                                                                                                          self.props[symbol]['qty'], self.props[symbol]['entry'], self.props[symbol]['SL'], self.price_decimals(symbol, self.props[symbol]['SL']-self.props[symbol]['entry'])), flush=True)
                    self.props[symbol]['qty'] = 0
                elif self.props[symbol]['qty'] < 0 and cp >= self.props[symbol]['SL'] and self.client.open_contracts(symbol) == 0:
                    self.props[symbol]['open'] = False
                    print("Short Position {} closed. {} Contracts from {} to {}. PnL in Points: {}".format(symbol,
                                                                                                           -self.props[symbol]['qty'], self.props[symbol]['entry'], self.props[symbol]['SL'], self.price_decimals(symbol, self.props[symbol]['entry']-self.props[symbol]['SL'])), flush=True)
                    self.props[symbol]['qty'] = 0

    def generate_id(self, symbol, _type='stop'):
        return "{}_{}_{}_{}".format(self.magic, symbol, str(uuid.uuid4().fields[-1])[:5], _type)

    def price_decimals(self, symbol, price):
        if symbol == 'XBTUSD':
            return round(price * 2) / 2
        elif symbol == 'ETHUSD':
            return round(price * 20) / 20
        else:
            print('{} is not a valid symbol for price decimal rounding.'.format(symbol))
            sys.exit()

    def calc_pos_size(self, symbol, sl_distance):
        acc_balance = self.client.acc_balance(symbol)
        ps = float(self.settings.symbols[symbol]['position_size'])
        if symbol == 'XBTUSD':
            btc_usd = self.client.last_current_price(symbol)
            lot_step = 100
            riskvalue = (float(acc_balance) / 100000000) * \
                float(btc_usd) * (ps/100)  # riskvalue in usd
            pos_size = int(
                round((math.floor(riskvalue / (sl_distance / btc_usd)) / lot_step)) * lot_step)
        elif symbol == 'ETHUSD':
            tickvalue = 100
            lot_step = 1
            riskvalue = acc_balance * (ps/100)
            pos_size = int(
                round((math.floor(riskvalue / (sl_distance * tickvalue)) / lot_step)) * lot_step)
        else:
            print('Symbol not supported')
            sys.exit()
        if pos_size < lot_step:
            return lot_step
        else:
            return pos_size

    def stoploss_order(self, symbol, execPrice='LastPrice'):
        try:
            self.props[symbol]['stoploss_id'] = self.generate_id(
                symbol, _type='SL')
            self.save_open_id(
                symbol, self.props[symbol]['stoploss_id'], stop_type='stoploss')
            return self.client.client.Order.Order_new(
                symbol=symbol, ordType='Stop', clOrdID=self.props[symbol]['stoploss_id'], orderQty=-self.props[symbol]['qty'],
                stopPx=self.price_decimals(symbol, self.props[symbol]['SL']), execInst=execPrice
            ).result()
        except HTTPBadRequest as e:
            print("Error placing stoploss order.", e, flush=True)

    def order_cancel(self, symbol, orderId):
        try:
            order_res = self.client.client.Order.Order_cancel(
                clOrdID=orderId
            ).result()
            self.props[symbol]['wait_stop'] = False
            return order_res
        except (HTTPBadRequest, HTTPNotFound) as e:
            print("Error canceling order.", e, flush=True)

    def stoploss_cancel(self, symbol):
        self.order_cancel(symbol, self.props[symbol]['stoploss_id'])

    def stoporder_cancel(self, symbol):
        self.order_cancel(symbol, self.props[symbol]['stop_id'])

    def bracket_stop_order(self, symbol, orderQty, entry, sl, tp=0):
        try:
            self.props[symbol]['stop_id'] = self.generate_id(symbol)
            order_res = self.client.client.Order.Order_new(
                symbol=symbol, ordType='Stop', clOrdID=self.props[symbol]['stop_id'], orderQty=orderQty,
                stopPx=self.price_decimals(symbol, entry), execInst='LastPrice'
            ).result()
            self.save_open_id(symbol, self.props[symbol]['stop_id'])
            self.props[symbol]['qty'] = orderQty
            self.props[symbol]['wait_stop'] = True
            self.props[symbol]['SL'] = sl
            self.props[symbol]['TP'] = tp
            self.props[symbol]['entry'] = entry
            return order_res
        except HTTPBadRequest as e:
            print("Error placing bracket stop order in {}.".format(
                symbol), e, flush=True)

    def bracket_market_order(self, symbol, orderQty, sl, cp, tp=0):
        try:
            entry = self.market_order(symbol, orderQty)
            self.props[symbol]['SL'] = sl
            self.props[symbol]['TP'] = tp
            time.sleep(5)
            stoploss = self.stoploss_order(symbol)
            self.props[symbol]['qty'] = orderQty
            self.props[symbol]['entry'] = cp
            return entry, stoploss
        except HTTPBadRequest as e:
            print("Error placing bracket market order in {}.".format(
                symbol), e, flush=True)

    def market_order(self, symbol, orderQty):
        try:
            order_res = self.client.client.Order.Order_new(
                symbol=symbol, ordType='Market', orderQty=orderQty
            ).result()
            self.props[symbol]['qty'] = orderQty
            self.props[symbol]['open'] = True
            self.props[symbol]['wait_stop'] = False
            return order_res
        except HTTPBadRequest as e:
            print("Error placing market order.", e, flush=True)

    def modifiy_stop(self, symbol, newStopLoss):
        try:
            # could be clOrdID, not sure difference
            res = self.client.client.Order.Order_amend(
                origClOrdID=self.props[symbol]['stoploss_id'], stopPx=self.price_decimals(symbol,
                                                                                          newStopLoss)
            ).result()
            self.props[symbol]['SL'] = newStopLoss
            return res
        except HTTPBadRequest as e:
            if 'Invalid amend' in str(e):
                return 'noValChanged'
            elif 'Invalid origClOrdID' in str(e):
                return 'invalidID'
            else:
                return str(e)

    def close(self, symbol, qty=0):
        if not qty:
            qty = -self.props[symbol]['qty']
        if self.props[symbol]['open']:
            # note: have to delete stop too
            try:
                close = self.client.client.Order.Order_new(
                    symbol=symbol, ordType='Market', execInst='Close', orderQty=qty
                ).result()
                self.stoploss_cancel(symbol)
                return close
            except HTTPBadRequest as e:
                print("Error closing order in {}.".format(symbol), e, flush=True)
        else:
            return False


api_key = ''
api_secret = ''

# binance live account
# api_key = ''
# api_secret = ''

client = BinanceClient(api_key=api_key, api_secret=api_secret, testnet=True)

print(client.get_histories())

# balance = client.client.futures_coin_account_balance()
# account = client.client.futures_account()

# print(balance)
# print(account)


# import math

# sym = 'BTCUSDT'

# symbol_info = client.client.get_symbol_info(sym)
# step_size = 0.0
# for f in symbol_info['filters']:
#     if f['filterType'] == 'LOT_SIZE':
#         step_size = float(f['stepSize'])

# quantity = 5

# precision = int(round(-math.log(step_size, 10), 0))
# quantity = float(round(quantity, precision))

# order = client.client.futures_create_order(symbol=sym, side='BUY', type='MARKET', quantity=quantity)

# klines = client.client.get_historical_klines("BTCUSDT", binanceClient.KLINE_INTERVAL_30MINUTE, "1 day ago UTC")

# print(klines)

'''
{
    'XBTUSD': 
        (
            [
                {'timestamp': datetime.datetime(2021, 9, 24, 15, 15, tzinfo=tzutc()),
                'symbol': 'XBTUSD',
                'open': 42215.5, 'high': 42215.5, 'low': 42150.0, 'close': 42205.0,
                'trades': 16, 'volume': 1900, 'vwap': 42196.9416, 'lastSize': 100, 'turnover': 4502700, 'homeNotional': 0.045027, 'foreignNotional': 1900.0},
                {'timestamp': datetime.datetime(2021, 9, 24, 15, 10, tzinfo=tzutc()),
                'symbol': 'XBTUSD',
                'open': 42171.0, 'high': 42271.0, 'low': 42094.0, 'close': 42215.5, 'trades': 31, 'volume': 7200, 'vwap': 42178.0758, 'lastSize': 100, 'turnover': 17070498, 'homeNotional': 0.17070498, 'foreignNotional': 7200.0}, {'timestamp': datetime.datetime(2021, 9, 24, 15, 5, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 42354.5, 'high': 42354.5, 'low': 42171.0, 'close': 42171.0, 'trades': 18, 'volume': 5300, 'vwap': 42337.361, 'lastSize': 100, 'turnover': 12518507, 'homeNotional': 0.12518507, 'foreignNotional': 5300.0}, {'timestamp': datetime.datetime(2021, 9, 24, 15, 0, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 42363.0, 'high': 42363.0, 'low': 42187.0, 'close': 42354.5, 'trades': 15, 'volume': 7600, 'vwap': 42351.5261, 'lastSize': 100, 'turnover': 17945096, 'homeNotional': 0.17945095999999994, 'foreignNotional': 7600.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 55, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 42247.5, 'high': 42363.0, 'low': 42169.5, 'close': 42363.0, 'trades': 23, 'volume': 10100, 'vwap': 42292.0605, 'lastSize': 200, 'turnover': 23881640, 'homeNotional': 0.23881639999999996, 'foreignNotional': 10100.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 50, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 42363.0, 'high': 42363.0, 'low': 42168.5, 'close': 42247.5, 'trades': 30, 'volume': 10000, 'vwap': 42295.1014, 'lastSize': 100, 'turnover': 23643416, 'homeNotional': 0.23643415999999998, 'foreignNotional': 10000.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 45, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 42305.0, 'high': 42363.0, 'low': 42307.5, 'close': 42363.0, 'trades': 18, 'volume': 5300, 'vwap': 42342.2012, 'lastSize': 200, 'turnover': 12517109, 'homeNotional': 0.12517109, 'foreignNotional': 5300.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 40, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 42096.0, 'high': 42351.0, 'low': 42051.0, 'close': 42305.0, 'trades': 40, 'volume': 72700, 'vwap': 42122.2894, 'lastSize': 800, 'turnover': 172592829, 'homeNotional': 1.7259282900000006, 'foreignNotional': 72700.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 35, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 41998.5, 'high': 42096.0, 'low': 41987.5, 'close': 42096.0, 'trades': 41, 'volume': 44100, 'vwap': 42055.8586, 'lastSize': 100, 'turnover': 104860813, 'homeNotional': 1.0486081299999999, 'foreignNotional': 44100.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 30, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 41964.5, 'high': 41998.5, 'low': 41909.0, 'close': 41998.5, 'trades': 23, 'volume': 6500, 'vwap': 41978.7084, 'lastSize': 1900, 'turnover': 15484098, 'homeNotional': 0.15484098, 'foreignNotional': 6500.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 25, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 41911.0, 'high': 41964.5, 'low': 41908.5, 'close': 41964.5, 'trades': 33, 'volume': 10600, 'vwap': 41934.1717, 'lastSize': 200, 'turnover': 25277720, 'homeNotional': 0.2527772, 'foreignNotional': 10600.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 20, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 41947.0, 'high': 41948.0, 'low': 41886.5, 'close': 41911.0, 'trades': 29, 'volume': 10300, 'vwap': 41937.1614, 'lastSize': 100, 'turnover': 24560617, 'homeNotional': 0.24560616999999998, 'foreignNotional': 10300.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 15, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 41946.0, 'high': 41947.0, 'low': 41887.5, 'close': 41947.0, 'trades': 20, 'volume': 2100, 'vwap': 41931.5342, 'lastSize': 100, 'turnover': 5008181, 'homeNotional': 0.050081809999999984, 'foreignNotional': 2100.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 10, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 41879.0, 'high': 41947.0, 'low': 41879.0, 'close': 41946.0, 'trades': 21, 'volume': 5000, 'vwap': 41933.2925, 'lastSize': 100, 'turnover': 11923711, 'homeNotional': 0.11923711000000001, 'foreignNotional': 5000.0}, {'timestamp': datetime.datetime(2021, 9, 24, 14, 5, tzinfo=tzutc()), 'symbol': 'XBTUSD', 'open': 41703.5, 'high': 41879.0, 'low': 41662.5, 'close': 41879.0, 'trades': 47, 'volume': 25400, 'vwap': 41769.8731, 'lastSize': 100, 'turnover': 60809419, 'homeNotional': 0.6080941899999998, 'foreignNotional': 25400.0}],
            < bravado.requests_client.RequestsResponseAdapter object at 0x7fb5fb583cd0 >
        )
}
'''
